// blockfind.c
//
// finds haplotype blocks based on a given measure.
// currently measure is: confidence limits on d prime
//
// usage: blockfind -d datafile [-f blockfile -o outputfile -s selectalg -p LDpercentage -t LDthreshold -b minblocksize]

#include <stdio.h>
#include <stdlib.h>
#include "LDvals.h"
#include "defs.h"

int main(int argc, char *argv[]) {

  char *LDfilename = "foo.out";
  double LDstrongthreshold = THRESHOLD;
  double LDpercentage = CUTOFF; // usually .95
  int minblocksize = 500;
  char *blockselectalg = "maxsize";
  char *datafile = "plot.dat";
  char *blockfile = "block.dat";


  int nextarg = 1; // if the argument has no .. arguments, then we can see if the next char is another argument.

  int c;

  int wrongargs = argc; // if == 1, we print usage.


  while (--argc > 0 && (*++argv)[0] == '-') {
    nextarg = 1;
    while ((c = *++argv[0]) && nextarg)
      switch (c) {

      case 'f':
	blockfile = (*++argv);
	printf ("the blockfile specified is %s\n", blockfile);
	nextarg = 0;
	argc--;
	break;

      case 'd':
	LDfilename = (*++argv);
	printf ("the filename specified is %s\n", LDfilename);
	nextarg = 0;
	argc--;
	break;

      case 'o':
	datafile = (*++argv);
	printf ("the datafile specified is %s\n", datafile);
	nextarg = 0;
	argc--;
	break;

      case 's':
	blockselectalg = (*++argv);
	printf ("the block select alg specified is %s\n", blockselectalg);
	nextarg = 0;
	argc--;
	break;

      case 'p': // LDpercentage;
	LDpercentage = atof(*++argv);
	printf ("the percentage specified is %.3f\n", LDpercentage);
	nextarg = 0;
	argc--;
	break;


      case 't': // LDstrongthreshold;
	LDstrongthreshold = atof(*++argv);
	printf ("the threshold specified is %.3f\n", LDstrongthreshold);
	nextarg = 0;
	argc--;
	break;

      case 'b': // minblocksize
	minblocksize = atoi(*++argv);
	printf ("the min blocksize specified is %d\n", minblocksize);
	nextarg = 0;
	argc--;
	break;

      default:
	printf("blockfind: illegal option %c\n", c);
	argc = 0;
	wrongargs = 1;
	break;
       
      }
  }
  if (wrongargs == 1) { // if there are no arguments
    printf("\nUsage: blockfind [-d datafile][-f blockfile][-o outputfile]\n\t\t[-s selectalg][-p LDpercentage][-t LDthreshold][-b minblocksize]\n");
    printf("(for some forgotten reason, you need to use at least one of these options, otherwise you'll just get this help message again.)\n\n");
    printf("if options not specified, default values are:\n");
    printf("\tdatafile: %s\n", LDfilename);
    printf("\tblockfile: %s\n", blockfile);
    printf("\toutputfile: %s\n", datafile);
    printf("\tselectalg: %s [i think this is obsolete]\n", blockselectalg);
    printf("\tpercentage: %.2f\n", LDpercentage);
    printf("\tLDthreshold: %.2f\n", LDstrongthreshold);
    printf("\tminblocksize: %d\n", minblocksize);
    printf("\nLDthreshold is currently obsolete because we are using confidence bounds. ");
    printf("Can be used to set threshold when other measures of LD are used.\n\n");
  }
  else { // what to do now.

    blockfind(LDfilename, LDstrongthreshold, LDpercentage, minblocksize, blockselectalg,
	      datafile, blockfile);
  }



}

//*************************************************************************

// getLongestBlock
//
// finds the longest block between start and end, up to MAXBLOCKLENGTH
// then recurses to find the largest block in the two remaining portions.
// 200 base space is allowed between blocks.
//
// this function currently has parameters from the altshuler paper [Gabriel et al.
// Science 296: 2225-2229]. when switching between populations, this section needs to
// be edited to change the distance limits. LDval.c needs to be edited to add in the
// different thresholds. a pain, i know.
//
int getLongestBlock(int start, int end, int blockarray[][2], int blockindex, double LDpercentage) {
 
  int i, dist;
  double percentage;
  int longest0 = -1, longest1 = -1, maxlength = 0;
  int tempcount = 0, longestblock, j, length = 0;
  //int tempvar;

  if (start < end) {

    for (i = start; i < end; i++) {
      longestblock = i;

      // let's look at every possible block.
      // MAXBLOCKLENGTH is set inside defs.h
      for (j = chooseSNP(i, end, MAXBLOCKLENGTH); 
	   (j >= i + 1) && (longestblock == i) ; j--) {
		
	if (paircheck(i, j)) {
	  dist = j - i;

	  // SPECIAL CASES
	  // for 2, 3, 4 marker blocks

	  // TWO MARKER BLOCK
	  if (dist == 1 && twocheck(i,j)) {
	    if (getSNPposition(j) - getSNPposition(i) < 10000)
	      longestblock = j;
	  }

	  // THREE MARKER BLOCK
	  else if (dist == 2) {
	    if (threecheck(i,i+1) && threecheck(i+1,j)) {
	      if (getSNPposition(j) - getSNPposition(i) < 20000)
		longestblock = j;
	    }
	  }

	  // FOUR MARKER BLOCK
	  else if (dist == 3) {
	    percentage = test(i,j);
	    if (percentage > LDpercentage) {
	      if (getSNPposition(j) - getSNPposition(i) < 30000)
		longestblock = j;
	    }
	  }

	  // EVERYTHING ELSE
	  else if (dist > 3) {
	    percentage = test(i,j);
	    
	    if (percentage > LDpercentage) {
	      longestblock = j;
	    }
	  } 
	}
      } 
      
      length = getSNPposition(longestblock) - getSNPposition(i);
    
      if (length > maxlength) {
	maxlength = length;

	// if you want to see the blocks as they are selected, unselect this:
	//
	//      printf("the longest block is from %d to %d\n",
	//	     i, longestblock);

	longest0 = i;
	longest1 = longestblock;

      }
    }

    // you've found the longest block. save it and move on.
    if (maxlength > 0) {
      
      blockarray[blockindex][0] = longest0;
      blockarray[blockindex][1] = longest1;
      blockindex++;
    }

    // now recurse to find the next longest block.
    if (longest0 != -1)
      blockindex = getLongestBlock(start, belowgap(longest0), blockarray, blockindex, LDpercentage);
    if (longest1 != -1)
      blockindex = getLongestBlock(abovegap(longest1), end, blockarray, blockindex, LDpercentage);  
  }

  return blockindex;
}

//*************************************************************************

// blockfind
//
// this is what the main function calls.
// 
blockfind(char *LDfile, double LDstrongthreshold, double LDpercentage, int minblocksize, char *blockselectalg, char *datafile, char *blockfile) {

  int SNPcount;
  int longestblock, dist;
  
  int blockarray[MAXSNPS][2];
  int blockcount = 0;

  int i, j;
  double percentage;

  FILE *picfp;
  FILE *datfp;
  FILE *blockfp;

  // CHANGE THIS TO USE SINGLE LD VALUES INSTEAD OF CONFIDENCE LIMITS.
  // no guarantees that simply uncommenting this will give you the desired results..
  // i might have hardwired in something bad.
  //
  //  SNPcount = readLDvalues(LDfile);
  //
  SNPcount = readConfvalues(LDfile);

  // so this, i think, is unnecessary for confidence limits. but i'll leave it here anyways.
  setThreshold(LDstrongthreshold);

  // run the block finding algorithm.
  blockcount = getLongestBlock(0,SNPcount,blockarray,0, LDpercentage);


  if (blockcount > 0) {
  datfp = fopen(datafile, "a");
  //  fprintf(datfp, "TH\tPerc\tMinB\tSelect\tBlocks\tLength\tShort\tLong\tCover\n");
  fprintf(datfp, "%.3f\t%.3f\t%d\t%s\t",
	 LDstrongthreshold, LDpercentage, minblocksize, blockselectalg);  
  printstats(blockarray,blockcount, datfp);

  blockfp = fopen(blockfile, "w");
  printblocks(blockarray, blockcount, blockfp);

  }
  else
  printf("no blocks.\n");
}

//*************************************************************************

// printblocks
//
// prints out the blocks that you have to blockfp.
// just dumps the array into the file. no sorting.
// your file will have a header on it.
//
printblocks(int array[][2], int count, FILE *blockfp) {
  int i;
  fprintf(blockfp, "SNP1\tindex\tSNP2\tindex\n");

  for (i = 0; i < count; i++) {
    fprintf(blockfp, "%d\t%d\t%d\t%d\n", getSNPposition(array[i][0]),
	    array[i][0], getSNPposition(array[i][1]), array[i][1]);
  }

}


//*************************************************************************

// printstats
//
// prints to stdout the stats from that last run.
//

printstats(int array[][2], int count, FILE *datfp) {
  int i;
  int totallength = 0;
  double average = 0, coverage;
  int size, longest = 0, shortest = 0;

  size = getSNPposition(array[0][1]) - getSNPposition(array[0][0]);
  longest = size;
  shortest = size;

  for (i = 0; i < count; i++) {
    size = getSNPposition(array[i][1]) - getSNPposition(array[i][0]);
    totallength += size;

    if (size > longest) longest = size;
    else if (size < shortest) shortest = size;
  }
  
  average = (double)totallength / (double) count;
  coverage = (double)totallength / (double)LENGTH;

  printf("Total number of blocks: %d\n", count);
  printf("Average length of blocks: %.3f\n", average);
  printf("Shortest block: %d\n", shortest);
  printf("Longest block: %d\n", longest);
  printf("Percent coverage: %.3f\n", coverage);

  fprintf(datfp, "%d\t%.3f\t%d\t%d\t%.3f\n",
	 count, average, shortest, longest, coverage);
}


//*************************************************************************
// END of blockfind.c
//
// lots of obsolete functions were deleted from here.
// they can be found in my directory if necessary:
// /gauguin/catfoo/trytwo/blockfind.c
// 
